@@ -17,11 +17,22 @@ function _isSharing(sd) {
 function _isRecording(sd) {
 	return !!sd && 'SCREEN' === sd.type && sd.activities.includes(REC_ACTIVITY);
 }
+function _hasActivity(sd, act) {
+	return !!sd && sd.activities.includes(act);
+}
 function _hasMic(sd) {
-	return !sd || sd.activities.includes(MIC_ACTIVITY);
+	if (!sd) {
+		return true;
+	}
+	const enabled = sd.micEnabled !== false;
+	return sd.activities.includes(MIC_ACTIVITY) && enabled;
 }
 function _hasCam(sd) {
-	return !sd || sd.activities.includes(CAM_ACTIVITY);
+	if (!sd) {
+		return true;
+	}
+	const enabled = sd.camEnabled !== false;
+	return sd.activities.includes(CAM_ACTIVITY) && enabled;
 }
 function _hasVideo(sd) {
 	return _hasCam(sd) || _isSharing(sd) || _isRecording(sd);
@@ -184,11 +195,10 @@ function _cleanStream(stream) {
 		stream.getTracks().forEach(track => track.stop());
 	}
 }
-function _cleanPeer(peer) {
-	if (!!peer) {
-		peer.cleaned = true;
+function _cleanPeer(rtcPeer) {
+	if (!!rtcPeer) {
 		try {
-			const pc = peer.peerConnection;
+			const pc = rtcPeer.pc;
 			if (!!pc) {
 				pc.getSenders().forEach(sender => {
 					try {
@@ -208,22 +218,8 @@ function _cleanPeer(peer) {
 						OmUtil.log('Failed to clean receiver' + e);
 					}
 				});
-				pc.onconnectionstatechange = null;
-				pc.ontrack = null;
-				pc.onremovetrack = null;
-				pc.onremovestream = null;
-				pc.onicecandidate = null;
-				pc.oniceconnectionstatechange = null;
-				pc.onsignalingstatechange = null;
-				pc.onicegatheringstatechange = null;
-				pc.onnegotiationneeded = null;
 			}
-			peer.dispose();
-			peer.removeAllListeners('icecandidate');
-			delete peer.generateOffer;
-			delete peer.processAnswer;
-			delete peer.processOffer;
-			delete peer.addIceCandidate;
+			rtcPeer.dispose();
 		} catch(e) {
 			//no-op
 		}
@@ -236,28 +232,24 @@ function _setPos(v, pos) {
 }
 function _askPermission(callback) {
 	const perm = $('#ask-permission');
-	if (undefined === perm.dialog('instance')) {
-		perm.data('callbacks', []).dialog({
-			appendTo: '.room-block .room-container'
-			, dialogClass: "ask-video-play-permission"
-			, autoOpen: false
-			, buttons: [
-				{
-					text: perm.data('btn-ok')
-					, click: function() {
-						while (perm.data('callbacks').length > 0) {
-							perm.data('callbacks').pop()();
-						}
-						$(this).dialog('close');
-					}
-				}
-			]
-		});
-	}
-	if (!perm.dialog('isOpen')) {
-		perm.dialog('open');
-	}
-	perm.data('callbacks').push(callback);
+	$('.sidebar').confirmation({
+		title: perm.attr('title')
+		, placement: Settings.isRtl ? 'right' : 'left'
+		, singleton: true
+		, rootSelector: '.sidebar'
+		, html: true
+		, content: perm.html()
+		, buttons: [{
+			class: 'btn btn-sm btn-warning'
+			, label: perm.data('btn-ok')
+			, value: perm.data('btn-ok')
+			, onClick: function() {
+				callback();
+				$('.sidebar').confirmation('dispose');
+			}
+		}]
+	});
+	$('.sidebar').confirmation('show');
 }
 function _disconnect(node) {
 	try {
@@ -291,10 +283,24 @@ function _highlight(el, clazz, count) {
 		next();
 	});
 }
+function _playSrc(_video, _stream, mute) {
+	if (_stream && _video) {
+		_video.srcObject = _stream;
+		if (_video.paused) {
+			_video.play().then(() => _video.muted = mute).catch(err => {
+				if ('NotAllowedError' === err.name) {
+					_askPermission(() => _video.play().then(() => _video.muted = mute));
+				}
+			});
+		}
+	}
+}
 
 module.exports = {
 	VIDWIN_SEL: VIDWIN_SEL
 	, VID_SEL: VID_SEL
+	, CAM_ACTIVITY: CAM_ACTIVITY
+	, MIC_ACTIVITY: MIC_ACTIVITY
 
 	, getVid: _getVid
 	, isSharing: _isSharing
@@ -302,6 +308,7 @@ module.exports = {
 	, hasMic: _hasMic
 	, hasCam: _hasCam
 	, hasVideo: _hasVideo
+	, hasActivity: _hasActivity
 	, getRects: _getRects
 	, getPos: _getPos
 	, container: _container
@@ -311,7 +318,7 @@ module.exports = {
 	, cleanPeer: _cleanPeer
 	, addIceServers: function(opts, m) {
 		if (m && m.iceServers && m.iceServers.length > 0) {
-			opts.configuration = {iceServers: m.iceServers};
+			opts.iceServers = m.iceServers;
 		}
 		return opts;
 	}
@@ -320,4 +327,5 @@ module.exports = {
 	, disconnect: _disconnect
 	, sharingSupported: _sharingSupported
 	, highlight: _highlight
+	, playSrc: _playSrc
 };
