@@ -20,7 +20,6 @@ package org.apache.openmeetings.web.room;
 
 import static de.agilecoders.wicket.core.markup.html.bootstrap.dialog.Modal.BUTTON_MARKUP_ID;
 import static java.time.Duration.ZERO;
-import static org.apache.openmeetings.core.remote.KurentoHandler.activityAllowed;
 import static org.apache.openmeetings.core.util.ChatWebSocketHelper.ID_USER_PREFIX;
 import static org.apache.openmeetings.db.entity.calendar.Appointment.allowedStart;
 import static org.apache.openmeetings.util.OmFileHelper.EXTENSION_PDF;
@@ -38,8 +37,6 @@ import java.util.Set;
 import java.util.stream.Stream;
 
 import org.apache.commons.text.StringEscapeUtils;
-import org.apache.openmeetings.core.remote.KurentoHandler;
-import org.apache.openmeetings.core.remote.StreamProcessor;
 import org.apache.openmeetings.core.util.WebSocketHelper;
 import org.apache.openmeetings.db.dao.calendar.AppointmentDao;
 import org.apache.openmeetings.db.dao.file.FileItemDao;
@@ -100,6 +97,8 @@ import org.apache.wicket.spring.injection.annot.SpringBean;
 import org.apache.wicket.util.resource.FileResourceStream;
 import org.apache.wicket.util.resource.IResourceStream;
 import org.apache.wicket.util.string.Strings;
+import org.apache.openmeetings.mediaserver.KurentoHandler;
+import org.apache.openmeetings.mediaserver.StreamProcessor;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 
@@ -144,7 +143,6 @@ public class RoomPanel extends BasePanel {
 		}
 	}
 	private final Room r;
-	private final boolean interview;
 	private final WebMarkupContainer room = new WebMarkupContainer("roomContainer");
 	private final AbstractDefaultAjaxBehavior roomEnter = new AbstractDefaultAjaxBehavior() {
 		private static final long serialVersionUID = 1L;
@@ -158,7 +156,7 @@ public class RoomPanel extends BasePanel {
 					.put("uid", c.getUid())
 					.put("userId", c.getUserId())
 					.put("rights", c.toJson(true).getJSONArray("rights"))
-					.put("interview", interview)
+					.put("interview", r.isInterview())
 					.put("audioOnly", r.isAudioOnly())
 					.put("allowRecording", r.isAllowRecording())
 					.put("questions", r.isAllowUserQuestions())
@@ -190,6 +188,7 @@ public class RoomPanel extends BasePanel {
 				showIdeaAlert(target, getString("1315"));
 			}
 			wb.update(target);
+			jsInited = true;
 		}
 
 		private CharSequence sendClientsOnInit() {
@@ -211,7 +210,7 @@ public class RoomPanel extends BasePanel {
 			if (streams.length() > 0) {
 				sb.append("VideoManager.play(").append(streams).append(", ").append(kHandler.getTurnServers(getClient())).append(");");
 			}
-			if (interview && streamProcessor.recordingAllowed(getClient())) {
+			if (r.isInterview() && streamProcessor.recordingAllowed(getClient())) {
 				sb.append("WbArea.setRecEnabled(true);");
 			}
 			if (!Strings.isEmpty(sb)) {
@@ -257,6 +256,7 @@ public class RoomPanel extends BasePanel {
 	};
 	Component eventDetail = new WebMarkupContainer(EVENT_DETAILS_ID).setVisible(false);
 	private boolean avInited = false;
+	private boolean jsInited = false;
 
 	@SpringBean
 	private ClientManager cm;
@@ -278,8 +278,7 @@ public class RoomPanel extends BasePanel {
 	public RoomPanel(String id, Room r) {
 		super(id);
 		this.r = r;
-		this.interview = Room.Type.INTERVIEW == r.getType();
-		this.wb = interview ? new InterviewWbPanel("whiteboard", this) : new WbPanel("whiteboard", this);
+		this.wb = r.isInterview() ? new InterviewWbPanel("whiteboard", this) : new WbPanel("whiteboard", this);
 	}
 
 	public void startDownload(IPartialPageRequestHandler handler, String type, String fuid) {
@@ -298,7 +297,7 @@ public class RoomPanel extends BasePanel {
 		room.setOutputMarkupPlaceholderTag(true);
 		room.add(menu = new RoomMenuPanel("menu", this));
 		room.add(AttributeModifier.append("data-room-id", r.getId()));
-		if (interview) {
+		if (r.isInterview()) {
 			room.add(new WebMarkupContainer("wb-area").add(wb));
 		} else {
 			Droppable<BaseFileItem> wbArea = new Droppable<>("wb-area") {
@@ -314,8 +313,7 @@ public class RoomPanel extends BasePanel {
 				@Override
 				public void onDrop(AjaxRequestTarget target, Component component) {
 					Object o = component.getDefaultModelObject();
-					if (wb.isVisible() && o instanceof BaseFileItem) {
-						BaseFileItem f = (BaseFileItem)o;
+					if (wb.isVisible() && o instanceof BaseFileItem f) {
 						if (sidebar.getFilesPanel().isSelected(f)) {
 							for (Entry<String, BaseFileItem> e : sidebar.getFilesPanel().getSelected().entrySet()) {
 								wb.sendFileToWb(e.getValue(), false);
@@ -420,11 +418,13 @@ public class RoomPanel extends BasePanel {
 
 	@Override
 	public void onEvent(IEvent<?> event) {
+		if (!jsInited) {
+			return;
+		}
 		Client curClient = getClient();
 		if (curClient != null && event.getPayload() instanceof WebSocketPushPayload) {
 			WebSocketPushPayload wsEvent = (WebSocketPushPayload) event.getPayload();
-			if (wsEvent.getMessage() instanceof RoomMessage) {
-				RoomMessage m = (RoomMessage)wsEvent.getMessage();
+			if (wsEvent.getMessage() instanceof RoomMessage m) {
 				IPartialPageRequestHandler handler = wsEvent.getHandler();
 				switch (m.getType()) {
 					case POLL_CREATED:
@@ -454,35 +454,35 @@ public class RoomPanel extends BasePanel {
 						roomClosed.show(handler);
 						break;
 					case REQUEST_RIGHT_MODERATOR:
-						sidebar.addActivity(new Activity((TextRoomMessage)m, Activity.Type.reqRightModerator), handler);
+						sidebar.addActivity(new Activity((TextRoomMessage)m, Activity.Type.REQ_RIGHT_MODERATOR), handler);
 						break;
 					case REQUEST_RIGHT_PRESENTER:
-						sidebar.addActivity(new Activity((TextRoomMessage)m, Activity.Type.reqRightPresenter), handler);
+						sidebar.addActivity(new Activity((TextRoomMessage)m, Activity.Type.REQ_RIGHT_PRESENTER), handler);
 						break;
 					case REQUEST_RIGHT_WB:
-						sidebar.addActivity(new Activity((TextRoomMessage)m, Activity.Type.reqRightWb), handler);
+						sidebar.addActivity(new Activity((TextRoomMessage)m, Activity.Type.REQ_RIGHT_WB), handler);
 						break;
 					case REQUEST_RIGHT_SHARE:
-						sidebar.addActivity(new Activity((TextRoomMessage)m, Activity.Type.reqRightShare), handler);
+						sidebar.addActivity(new Activity((TextRoomMessage)m, Activity.Type.REQ_RIGHT_SHARE), handler);
 						break;
 					case REQUEST_RIGHT_REMOTE:
-						sidebar.addActivity(new Activity((TextRoomMessage)m, Activity.Type.reqRightRemote), handler);
+						sidebar.addActivity(new Activity((TextRoomMessage)m, Activity.Type.REQ_RIGHT_REMOTE), handler);
 						break;
 					case REQUEST_RIGHT_A:
-						sidebar.addActivity(new Activity((TextRoomMessage)m, Activity.Type.reqRightA), handler);
+						sidebar.addActivity(new Activity((TextRoomMessage)m, Activity.Type.REQ_RIGHT_A), handler);
 						break;
 					case REQUEST_RIGHT_AV:
-						sidebar.addActivity(new Activity((TextRoomMessage)m, Activity.Type.reqRightAv), handler);
+						sidebar.addActivity(new Activity((TextRoomMessage)m, Activity.Type.REQ_RIGHT_AV), handler);
 						break;
 					case REQUEST_RIGHT_MUTE_OTHERS:
-						sidebar.addActivity(new Activity((TextRoomMessage)m, Activity.Type.reqRightMuteOthers), handler);
+						sidebar.addActivity(new Activity((TextRoomMessage)m, Activity.Type.REQ_RIGHT_MUTE_OTHERS), handler);
 						break;
 					case ACTIVITY_REMOVE:
 						sidebar.removeActivity(((TextRoomMessage)m).getText(), handler);
 						break;
 					case HAVE_QUESTION:
 						if (curClient.hasRight(Room.Right.MODERATOR) || getUserId().equals(m.getUserId())) {
-							sidebar.addActivity(new Activity((TextRoomMessage)m, Activity.Type.haveQuestion), handler);
+							sidebar.addActivity(new Activity((TextRoomMessage)m, Activity.Type.REQ_RIGHT_HAVE_QUESTION), handler);
 						}
 						break;
 					case KICK:
@@ -569,13 +569,13 @@ public class RoomPanel extends BasePanel {
 			handler.appendJavaScript("Room.addClient(["
 					+ c.toJson(self).toString(new NullStringer()) + "]);");
 		}
-		sidebar.addActivity(new Activity(m, Activity.Type.roomEnter), handler);
+		sidebar.addActivity(new Activity(m, Activity.Type.ROOM_ENTER), handler);
 	}
 
 	private void onRoomExit(TextRoomMessage m, IPartialPageRequestHandler handler) {
 		String uid = m.getText();
 		sidebar.update(handler);
-		sidebar.addActivity(new Activity(m, Activity.Type.roomExit), handler);
+		sidebar.addActivity(new Activity(m, Activity.Type.ROOM_EXIT), handler);
 		handler.appendJavaScript("Room.removeClient('" + uid + "'); Chat.removeTab('" + ID_USER_PREFIX + m.getUserId() + "');");
 	}
 
@@ -626,7 +626,7 @@ public class RoomPanel extends BasePanel {
 
 	private void updateInterviewRecordingButtons(IPartialPageRequestHandler handler) {
 		Client curClient = getClient();
-		if (interview && curClient.hasRight(Right.MODERATOR)) {
+		if (r.isInterview() && curClient.hasRight(Right.MODERATOR)) {
 			if (streamProcessor.isRecording(r.getId())) {
 				handler.appendJavaScript("if (typeof(WbArea) === 'object') {WbArea.setRecStarted(true);}");
 			} else if (streamProcessor.recordingAllowed(getClient())) {
@@ -675,8 +675,8 @@ public class RoomPanel extends BasePanel {
 
 	@Override
 	public void cleanup(IPartialPageRequestHandler handler) {
-		if (eventDetail instanceof EventDetailDialog) {
-			((EventDetailDialog)eventDetail).close(handler);
+		if (eventDetail instanceof EventDetailDialog evtDialog) {
+			evtDialog.close(handler);
 		}
 		handler.add(getBasePage().getHeader().setVisible(true), getMainPanel().getTopControls().setVisible(true));
 		if (r.isHidden(RoomElement.CHAT)) {
@@ -758,12 +758,14 @@ public class RoomPanel extends BasePanel {
 		for (Right right : rights) {
 			client.deny(right);
 		}
-		if (client.hasActivity(Client.Activity.AUDIO) && !client.hasRight(Right.AUDIO)) {
-			client.remove(Client.Activity.AUDIO);
-		}
-		if (client.hasActivity(Client.Activity.VIDEO) && !client.hasRight(Right.VIDEO)) {
-			client.remove(Client.Activity.VIDEO);
-		}
+		client.getCamStreams().forEach(sd -> {
+			if (sd.has(Client.Activity.AUDIO) && !client.hasRight(Right.AUDIO)) {
+				sd.remove(Client.Activity.AUDIO);
+			}
+			if (sd.has(Client.Activity.VIDEO) && !client.hasRight(Right.VIDEO)) {
+				sd.remove(Client.Activity.VIDEO);
+			}
+		});
 		rightsUpdated(client);
 	}
 
@@ -792,13 +794,13 @@ public class RoomPanel extends BasePanel {
 				if (!avInited) {
 					avInited = true;
 					if (Room.Type.CONFERENCE == r.getType()) {
-						if (!activityAllowed(c, Client.Activity.AUDIO, c.getRoom())) {
+						if (!c.isAllowed(Client.Activity.AUDIO)) {
 							c.allow(Room.Right.AUDIO);
 						}
-						if (!c.getRoom().isAudioOnly() && !activityAllowed(c, Client.Activity.VIDEO, c.getRoom())) {
+						if (!c.getRoom().isAudioOnly() && !c.isAllowed(Client.Activity.VIDEO)) {
 							c.allow(Room.Right.VIDEO);
 						}
-						streamProcessor.toggleActivity(c, c.getRoom().isAudioOnly()
+						streamProcessor.onToggleActivity(c, c.getRoom().isAudioOnly()
 								? Client.Activity.AUDIO
 								: Client.Activity.AUDIO_VIDEO);
 					}
@@ -842,7 +844,7 @@ public class RoomPanel extends BasePanel {
 	}
 
 	public boolean isInterview() {
-		return interview;
+		return r.isInterview();
 	}
 
 	private void createWaitModerator(final boolean autoopen) {
